from frasco import lazy_translate, request, flash, redirect
from frasco.ext import *
from frasco.utils import unknown_value
from frasco.models import as_transaction
from frasco.mail import send_mail
import stripe
import datetime
import time
import os
import json

from frasco.billing.eu_vat import get_exchange_rate, is_eu_country, model_rate_updated
from frasco.billing.invoicing import send_failed_invoice_mail

from .signals import *
from .webhook import webhook_blueprint
from .invoice import *
from .model import *


STRIPE_API_VERSION = "2020-03-02"
stripe.enable_telemetry = False


class FrascoStripe(Extension):
    name = "frasco_stripe"
    defaults = {"default_currency": None,
                "default_plan": None,
                "no_payment_redirect_to": None,
                "no_payment_message": None,
                "subscription_past_due_message": lazy_translate(
                    "We attempted to charge your credit card for your subscription but it failed."
                    "Please check your credit card details"),
                "debug_trial_period": None,
                "send_trial_will_end_email": True,
                "send_failed_invoice_mail": True,
                "invoice_ref_kwargs": {},
                "webhook_validate_event": False}

    def _init_app(self, app, state):
        stripe.api_key = state.require_option('api_key')
        stripe.api_version = STRIPE_API_VERSION
        state.Model = state.import_option('model')
        state.subscription_enabled = hasattr(state.Model, 'stripe_subscription_id')

        app.register_blueprint(webhook_blueprint)

        if has_extension("frasco_mail", app):
            app.extensions.frasco_mail.add_templates_from_package(__name__)

        if has_extension('frasco_eu_vat', app) and hasattr(state.Model, '__stripe_has_eu_vat__'):
            model_rate_updated.connect(lambda sender: sender.update_stripe_subscription_tax_rates(), weak=True)

        stripe_event_signal('customer_updated').connect(on_customer_updated_event)
        stripe_event_signal('customer_deleted').connect(on_customer_deleted_event)
        stripe_event_signal('payment_method_attached').connect(on_payment_method_event)
        stripe_event_signal('payment_method_detached').connect(on_payment_method_event)
        stripe_event_signal('payment_method_updated').connect(on_payment_method_event)
        stripe_event_signal('payment_method_card_automatically_updated').connect(on_payment_method_event)
        stripe_event_signal('invoice_payment_succeeded').connect(on_invoice_payment)
        stripe_event_signal('invoice_payment_failed').connect(on_invoice_payment)
        if state.subscription_enabled:
            stripe_event_signal('customer_subscription_created').connect(on_subscription_event)
            stripe_event_signal('customer_subscription_updated').connect(on_subscription_event)
            stripe_event_signal('customer_subscription_deleted').connect(on_subscription_event)
            stripe_event_signal('customer_subscription_trial_will_end').connect(on_trial_will_end)
            stripe_event_signal('invoice_created').connect(on_subscription_invoice_created)
        if hasattr(state.Model, '__stripe_has_eu_vat__'):
            stripe_event_signal('customer_tax_id_created').connect(on_tax_id_event)
            stripe_event_signal('customer_tax_id_updated').connect(on_tax_id_event)
            stripe_event_signal('customer_tax_id_deleted').connect(on_tax_id_event)


@as_transaction
def on_customer_updated_event(sender, stripe_event):
    state = get_extension_state('frasco_stripe')
    customer = stripe_event.data.object
    obj = state.Model.query_by_stripe_customer(customer.id).first()
    if obj:
        obj._update_stripe_customer(customer)


@as_transaction
def on_customer_deleted_event(sender, stripe_event):
    state = get_extension_state('frasco_stripe')
    customer = stripe_event.data.object
    obj = state.Model.query_by_stripe_customer(customer.id).first()
    if obj:
        obj._update_stripe_customer(False)


@as_transaction
def on_payment_method_event(sender, stripe_event):
    state = get_extension_state('frasco_stripe')
    source = stripe_event.data.object
    obj = state.Model.query_by_stripe_customer(source.customer).first()
    if obj:
        obj._update_stripe_customer()


@as_transaction
def on_tax_id_event(sender, stripe_event):
    state = get_extension_state('frasco_stripe')
    obj = state.Model.query_by_stripe_customer(stripe_event.data.object.customer).first()
    if obj:
        obj.update_from_stripe_eu_vat_number()


@as_transaction
def on_subscription_event(sender, stripe_event):
    state = get_extension_state('frasco_stripe')
    subscription = stripe_event.data.object
    obj = state.Model.query_by_stripe_customer(subscription.customer).first()
    if obj:
        obj._update_stripe_subscription()


@as_transaction
def on_subscription_invoice_created(sender, stripe_event):
    state = get_extension_state('frasco_stripe')
    invoice = stripe_event.data.object
    if not invoice.subscription:
        return
    obj = state.Model.query_by_stripe_customer(invoice.customer).first()
    if obj:
        obj.plan_has_invoice_items = False
        model_subscription_invoice_created.send(obj)


@as_transaction
def on_trial_will_end(sender, stripe_event):
    state = get_extension_state('frasco_stripe')
    subscription = stripe_event.data.object
    obj = state.Model.query_by_stripe_subscription(subscription.id).first()
    if obj and state.options['send_trial_will_end_email']:
        send_mail(getattr(obj, obj.__stripe_email_property__), 'stripe/trial_will_end.txt', obj=obj)


@as_transaction
def on_invoice_payment(sender, stripe_event):
    state = get_extension_state('frasco_stripe')
    invoice = stripe_event.data.object
    if not invoice.customer:
        return

    obj = state.Model.query_by_stripe_customer(invoice.customer).first()
    if not obj or invoice.total == 0:
        return

    if invoice.subscription:
        sub_obj = None
        if obj.stripe_subscription_id == invoice.subscription:
            sub_obj = obj
        else:
            sub_obj = state.Model.query_by_stripe_subscription(invoice.subscription).first()
        if sub_obj:
            sub_obj.plan_status = sub_obj.stripe_subscription.status
            sub_obj.update_last_stripe_subscription_invoice(invoice)

    invoice_payment.send(invoice)

    if has_extension('frasco_invoicing'):
        if invoice.paid:
            create_invoice_from_stripe(obj, invoice)
        elif not invoice.paid and state.options['send_failed_invoice_mail']:
            send_failed_invoice_mail(getattr(obj, obj.__stripe_email_property__), invoice)



import datetime


_translate_callback = None
_ntranslate_callback = None
_lazy_translate_callback = None
_format_datetime_callback = None
_format_date_callback = None
_format_time_callback = None


def set_translation_callbacks(translate=None, ntranslate=None, lazy_translate=None,\
    format_datetime=None, format_date=None, format_time=None):
    global _translate_callback, _ntranslate_callback, _lazy_translate_callback,\
        _format_datetime_callback, _format_date_callback, _format_time_callback
    _translate_callback = translate
    _ntranslate_callback = ntranslate
    _lazy_translate_callback = lazy_translate
    _format_datetime_callback = format_datetime
    _format_date_callback = format_date
    _format_time_callback = format_time


def translatable(string):
    # used to mark strings for extraction but no immediate translation
    return string


def translate(string, **kwargs):
    if _translate_callback:
        return _translate_callback(string, **kwargs)
    return string

_ = translate


def ntranslate(singular, plural, num, **kwargs):
    if _ntranslate_callback:
        return _ntranslate_callback(singular, plural, num, **kwargs)
    if num > 1:
        return plural % {'num': num}
    return singular % {'num': num}


def lazy_translate(string, **kwargs):
    if _lazy_translate_callback:
        return _lazy_translate_callback(string, **kwargs)
    from speaklater import make_lazy_string
    return make_lazy_string(translate, string, **kwargs)


def format_datetime(dt=None, format=None, **kwargs):
    if _format_datetime_callback:
        return _format_datetime_callback(dt, format, **kwargs)
    if not dt:
        dt = datetime.datetime.now()
    if not format:
        return dt.strftime("%Y-%m-%d %H:%M:%S")
    return dt.strftime(format)


def format_date(d=None, format=None, **kwargs):
    if _format_date_callback:
        return _format_date_callback(d, format, **kwargs)
    if not d:
        d = datetime.date.today()
    if not format:
        return d.strftime("%Y-%m-%d")
    return d.strftime(format)


def format_time(t=None, format=None, **kwargs):
    if _format_time_callback:
        return _format_time_callback(t, format, **kwargs)
    if not t:
        t = datetime.time()
    if not format:
        return t.isoformat()
    return t.strftime(format)



from flask import json
import speaklater
import datetime


class JSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, speaklater._LazyString):
            return o.value
        if isinstance(o, (set, map)):
            return list(o)
        if isinstance(o, datetime.time):
            return str(o)
        return json.JSONEncoder.default(self, o)



from contextlib import contextmanager
import requests
import tempfile


def split_backend_from_filename(filename):
    if '://' in filename:
        return filename.split('://', 1)
    return None, filename


class StorageBackend(object):
    default_options = None

    def __init__(self, name, options):
        self.name = name
        self.options = dict(self.default_options or {})
        self.options.update(options)

    def save(self, file, filename):
        raise NotImplementedError()

    def url_for(self, filename, **kwargs):
        raise NotImplementedError()

    def delete(self, filename):
        raise NotImplementedError()

    def open(self, filename):
        raise NotImplementedError()


class RemoteOpenStorageBackendMixin(object):
    @contextmanager
    def open(self, filename):
        with tempfile.TemporaryFile() as f:
            r = requests.get(self.url_for(filename))
            f.write(r.content)
            f.seek(0)
            yield f



from flask import g
from frasco.ext import *
from frasco.i18n import set_translation_callbacks
from flask_babel import (Babel, gettext, pgettext, ngettext, npgettext, lazy_gettext, lazy_pgettext, format_datetime, format_date,
                         format_time, format_currency as babel_format_currency)
from babel import Locale

from .cli import babel_cli
from .ctx import *
from .user import update_user_with_locale


class FrascoBabelState(ExtensionState):
    def __init__(self, *args):
        super(FrascoBabelState, self).__init__(*args)
        self.extract_dirs = []
        self.babel = None
    
    def add_extract_dir(self, path, jinja_dirs=None, jinja_exts=None, extractors=None):
        jinja_exts = jinja_exts or []
        jinja_exts.extend(self.options["extract_with_jinja_exts"])
        self.extract_dirs.append((path, jinja_dirs, jinja_exts, extractors))


class FrascoBabel(Extension):
    name = "frasco_babel"
    state_class = FrascoBabelState
    prefix_extra_options = "BABEL_"
    defaults = {"locales": ["en"],
                "default_locale": "en",
                "currencies": ["USD"],
                "default_currency": "USD",
                "currency_name_format": "{name} ({symbol})",
                "store_locale_in_session": True,
                "store_locale_in_user": False,
                "user_locale_column": "locale",
                "user_timezone_column": "timezone",
                "user_currency_column": "currency",
                "extract_locale_from_headers": True,
                "extract_locale_from_request": False,
                "always_add_locale_to_urls": True,
                "store_request_locale_in_session": False,
                "request_arg": "locale",
                "extractors": [],
                "extract_keywords": [],
                "extract_jinja_dirs": ["templates", "emails"],
                "extract_with_jinja_exts": ["jinja2.ext.autoescape", "jinja2.ext.with_",
                    "jinja2.ext.do", "jinja_layout.LayoutExtension", "jinja_macro_tags.LoadMacroExtension",
                    "jinja_macro_tags.CallMacroTagExtension", "jinja_macro_tags.JinjaMacroTagsExtension",
                    "jinja_macro_tags.HtmlMacroTagsExtension", "frasco.templating.FlashMessagesExtension"],
                "request_locale_arg_ignore_endpoints": ["static", "static_upload"],
                "compile_to_json": False,
                "compile_to_js": False,
                "js_catalog_varname": "LOCALE_%s_CATALOG",
                "babel_bin": "pybabel"}


    def _init_app(self, app, state):
        state.options.locales = set(state.options.locales)
        
        state.babel = Babel(app)
        state.babel.default_currency = state.options["default_currency"]
        state.babel.localeselector(detect_locale)
        state.babel.timezoneselector(detect_timezone)
        state.babel.currency_selector_func = detect_currency

        app.cli.add_command(babel_cli)

        set_translation_callbacks(translate=translate,
                                  ntranslate=ntranslate,
                                  lazy_translate=lazy_translate,
                                  format_datetime=format_datetime,
                                  format_date=format_date,
                                  format_time=format_time)

        from frasco.users.signals import user_signed_up
        user_signed_up.connect(lambda _, user: update_user_with_locale(user), weak=False)

        @app.url_value_preprocessor
        def extract_locale_from_values(endpoint, values):
            if state.options["extract_locale_from_request"] and values:
                values.pop(state.options["request_arg"], None)
        
        @app.url_defaults
        def add_locale_to_url_params(endpoint, values):
            if endpoint not in state.options["request_locale_arg_ignore_endpoints"] and \
              state.options["extract_locale_from_request"] and state.options['always_add_locale_to_urls'] and \
              state.options["request_arg"] not in values:
                values[state.options["request_arg"]] = get_locale().language

        @app.before_request
        def before_request():
            locale = get_locale()
            currency = get_currency()
            g.current_locale = locale.language
            g.current_timezone = get_timezone().zone
            g.current_currency = currency
            g.current_language = locale.display_name
            g.current_currency_name = state.options["currency_name_format"].format(
                code=currency,
                name=locale.currencies[currency],
                symbol=locale.currency_symbols[currency])

        app.jinja_env.globals.update(translate=gettext, ntranslate=ngettext, _=gettext,
            available_currencies=available_currencies, available_locales=available_locales)
        app.jinja_env.filters.update(datetimeformat=format_datetime, dateformat=format_date,
            timeformat=format_time, currencyformat=format_currency)


def translate(string, context=None):
    if context:
        return pgettext(context, string)
    return gettext(string)


def ntranslate(singular, plural, num, context=None):
    if context:
        return npgettext(context, singular, plural, num)
    return ngettext(singular, plural, num)


def lazy_translate(string, context=None):
    if context:
        return lazy_pgettext(context, string)
    return lazy_gettext(string)


def format_currency(number, format=None):
    return babel_format_currency(number, get_currency(), format)


def available_locales(english_name=False):
    locales = []
    for language in get_extension_state('frasco_babel').options["locales"]:
        locale = Locale(language)
        name = locale.english_name if english_name else locale.display_name
        locales.append((language, name))
    return locales


def available_currencies():
    currencies = []
    locale = get_locale()
    for currency in get_extension_state('frasco_babel').options["currencies"]:
        currencies.append((currency, locale.currencies[currency], locale.currency_symbols[currency]))
    return currencies
